# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Game engine that coordinates character actions to play a game.
"""

import random
from typing import Callable, Dict, List, Optional, Tuple

from attrs import define, field

from .game import Card, Game, GameMode, Player, PlayerColor, PlayerView
from .rules import Move, Rules
from .source import CharacterInputSource
from .util import CircularQueue


@define(slots=False)
class Character:
    # noinspection PyUnresolvedReferences
    """
    A character that plays a game, which could be a person or could be computer-driven.

    Attributes:
        name(str): The name of this character
        source(CharacterInputSource): The character input source from which moves are taken
    """

    name: str
    source: CharacterInputSource

    def choose_move(
        self, mode: GameMode, view: PlayerView, legal_moves: List[Move], evaluator: Callable[[PlayerView, Move], PlayerView]
    ) -> Move:
        """
        Choose the next move for a character via the user input source.

        Args:
            mode(GameMode): Game mode
            view(PlayerView): Player-specific view of the game
            legal_moves(Set[Move]): The set of legal moves
            evaluator(Callable[[PlayerView, Move], PlayerView]): Function to evaluate a move, returning new state

        Returns:
            Move: The character's next as chosen by the configured source
        """
        return self.source.choose_move(mode, view, legal_moves, evaluator)


@define
class Engine:
    # noinspection PyUnresolvedReferences
    """
    Game engine that coordinates character actions in a game.

    Normally, playing a game via an engine is as simple as::

        engine.start_game()
        while not engine.completed:
          state = engine.play_next()

    This synchronously plays a turn for each player, one after another, until
    the game is complete.  Other, more fine-grained methods exist if you need
    to structure game play differently for your purposes (for instance, to
    train a machine learning model or to play the game in an asynchronous
    event-driven environment).

    Attributes:
        mode(GameMode): The game mode
        characters(List[Character]): Characters playing the game
        first(PlayerColor): The first player, chosen randomly by default
    """

    mode: GameMode
    characters: List[Character]
    first: PlayerColor = field()
    _game: Game = field(init=False)
    _queue: CircularQueue[PlayerColor] = field(init=False)
    _rules: Rules = field(init=False)
    _map: Dict[PlayerColor, Character] = field(init=False)

    # noinspection PyUnresolvedReferences
    @first.default
    def _default_first(self) -> PlayerColor:
        return random.choice(list(PlayerColor)[: len(self.characters)])

    # noinspection PyUnresolvedReferences
    @_game.default
    def _default_game(self) -> Game:
        return Game(playercount=len(self.characters))

    @_queue.default
    def _default_queue(self) -> CircularQueue[PlayerColor]:
        return CircularQueue(list(PlayerColor)[: len(self.characters)], first=self.first)

    # noinspection PyUnresolvedReferences
    @_rules.default
    def _default_rules(self) -> Rules:
        return Rules(self.mode)

    # noinspection PyUnresolvedReferences
    @_map.default
    def _default_map(self) -> Dict[PlayerColor, Character]:
        index = 0
        result = {}
        for player in self._game.players.values():
            result[player.color] = self.characters[index]
            index += 1
        return result

    @property
    def players(self) -> int:
        """Number of players in the game."""
        return len(self.characters)

    @property
    def state(self) -> str:
        """String describing the state of the game."""
        if self.completed:
            return "Game completed"
        elif self.started:
            return "Game in progress"
        else:
            return "Game waiting to start"

    @property
    def game(self) -> Game:
        """A reference to the underlying game."""
        return self._game

    @property
    def started(self) -> bool:
        """Whether the game is started."""
        return self._game.started

    @property
    def completed(self) -> bool:
        """Whether the game is completed."""
        return self._game.completed

    @property
    def colors(self) -> Dict[PlayerColor, Character]:
        return self._map.copy()

    def winner(self) -> Optional[Tuple[Character, Player]]:
        """Return the winner of the game, as a tuple of (Character, Player)"""
        return (self._map[self._game.winner.color], self._game.winner) if self.completed else None  # type: ignore

    def reset(self) -> Game:
        """Reset game state."""
        self._game = self._default_game()
        return self._game

    def start_game(self) -> Game:
        """
        Start the game, returning game state.

        Returns:
            Game: Current state of the game.
        """
        self._rules.start_game(self._game)
        return self._game

    def next_turn(self) -> Tuple[PlayerColor, Character]:
        """
        Get the color and character for the next turn
        This will give you a different player each time you call it.
        """
        color = self._queue.next()
        character = self._map[color]
        return color, character

    def play_next(self) -> Game:
        """
        Play the next turn of the game, returning game state as of the end of the turn.

        Returns:
            Game: Current state of the game.
        """
        if self.completed:
            raise ValueError("Game is complete")

        saved = self._game.copy()
        try:
            color, character = self.next_turn()

            done = False
            while not done:
                view = self._game.create_player_view(color)
                move = self.choose_next_move(character, view)
                done = self.execute_move(color, move)

            return self._game
        except Exception as e:
            self._game = saved  # put back original so a failed call is idempotent
            raise e

    def draw(self) -> Card:
        """Draw a random card from the game's draw pile."""
        return self._game.deck.draw()

    def discard(self, card: Card) -> None:
        """Discard back to the game's discard pile."""
        self._game.deck.discard(card)

    def construct_legal_moves(self, view: PlayerView, card: Optional[Card] = None) -> Tuple[Optional[Card], List[Move]]:
        """Construct the legal moves based on a player view, using the passed-in card if provided."""
        card = card if card is not None else None if self.mode == GameMode.ADULT else self.draw()
        return card, self._rules.construct_legal_moves(view, card=card)

    def choose_next_move(self, character: Character, view: PlayerView) -> Move:
        """Choose the next move for a character based on a player view."""
        _, legal_moves = self.construct_legal_moves(view)
        move = character.choose_move(self.mode, view, legal_moves[:], Rules.evaluate_move)
        if move not in legal_moves:  # an illegal move is ignored and we choose randomly for the character
            self._game.track("Illegal move: a random legal move will be chosen", view.player)
            move = random.choice(legal_moves)
        return move

    def execute_move(self, color: PlayerColor, move: Move) -> bool:
        """Execute a move for a player, returning True if the player's turn is done."""
        player = self._game.players[color]
        if self.mode == GameMode.ADULT:
            return self._execute_move_adult(player, move)
        else:
            return self._execute_move_standard(player, move)

    def _execute_move_standard(self, player: Player, move: Move) -> bool:
        """Play the next turn under the rules for standard mode, returning True if the player's turn is done."""
        if not move.actions:
            self.discard(move.card)
            self._game.track("Turn is forfeit; discarded card %s" % move.card.cardtype.value, player, move.card)
            return True  # player's turn is done if they forfeit
        else:
            self._rules.execute_move(self._game, player, move)  # tracks history, potentially completes game
            self.discard(move.card)
            return self.completed or not self._rules.draw_again(move.card)  # player's turn is done unless they can draw again

    def _execute_move_adult(self, player: Player, move: Move) -> bool:
        """Play the next move under the rules for adult mode, returning True if the player's turn is done."""
        if not move.actions:
            player.hand.remove(move.card)
            self.discard(move.card)
            player.hand.append(self.draw())
            self._game.track("Turn is forfeit; discarded card %s" % move.card.cardtype.value, player, move.card)
            return True  # player's turn is done if they forfeit
        else:
            self._rules.execute_move(self._game, player, move)  # tracks history, potentially completes game
            player.hand.remove(move.card)
            self.discard(move.card)
            player.hand.append(self.draw())
            return self.completed or not self._rules.draw_again(move.card)  # player's turn is done unless they can draw again



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:
# pylint: disable=protected-access

from unittest.mock import MagicMock

import pendulum
import pytest
from pendulum.datetime import DateTime

from apologies.game import (
    BOARD_SQUARES,
    DECK_COUNTS,
    DECK_SIZE,
    PAWNS,
    SAFE_SQUARES,
    Card,
    CardType,
    Deck,
    Game,
    History,
    Pawn,
    Player,
    PlayerColor,
    PlayerView,
    Position,
)


class TestCard:
    def test_constructor(self):
        card = Card(0, CardType.CARD_12)
        assert card.id == 0
        assert card.cardtype == CardType.CARD_12


class TestDeck:
    def test_constructor(self):
        deck = Deck()

        assert len(deck._draw_pile.keys()) == DECK_SIZE
        assert len(deck._discard_pile.keys()) == 0

        cardcounts = {cardtype: 0 for cardtype in CardType}
        for card in deck._draw_pile.values():
            cardcounts[card.cardtype] += 1
        for cardtype in CardType:
            assert cardcounts[cardtype] == DECK_COUNTS[cardtype]

    def test_draw_and_discard(self):
        deck = Deck()

        # Check that we can draw the entire deck
        drawn = []
        for _ in range(DECK_SIZE):
            drawn.append(deck.draw())
        assert len(deck._draw_pile) == 0
        with pytest.raises(ValueError):
            deck.draw()

        # Discard one card and prove we can draw it
        card = drawn.pop()
        assert len(deck._discard_pile) == 0
        deck.discard(card)
        assert len(deck._discard_pile) == 1
        assert card is deck.draw()
        assert len(deck._discard_pile) == 0
        assert len(deck._draw_pile) == 0

        # Confirm that we're not allowed to discard the same card twice
        card = drawn.pop()
        deck.discard(card)
        with pytest.raises(ValueError):
            deck.discard(card)

        # Discard a few others and prove they can also be drawn
        deck.discard(drawn.pop())
        deck.discard(drawn.pop())
        deck.discard(drawn.pop())
        assert len(deck._discard_pile) == 4
        assert len(deck._draw_pile) == 0
        deck.draw()
        assert len(deck._discard_pile) == 0
        assert len(deck._draw_pile) == 3
        deck.draw()
        assert len(deck._discard_pile) == 0
        assert len(deck._draw_pile) == 2
        deck.draw()
        assert len(deck._discard_pile) == 0
        assert len(deck._draw_pile) == 1
        deck.draw()
        assert len(deck._discard_pile) == 0
        assert len(deck._draw_pile) == 0

        # Make sure the deck still gives us an error when empty
        with pytest.raises(ValueError):
            deck.draw()


# noinspection PyTypeHints
class TestPosition:
    def test_constructor(self):
        position = Position()
        assert position.start is True
        assert position.home is False
        assert position.safe is None
        assert position.square is None
        assert "%s" % position == "start"

    def test_copy(self):
        position = Position()
        assert position.start is True
        assert position.home is False
        position.safe = 9
        position.square = 13
        copy = position.copy()
        assert copy is not position and copy == position

    def test_move_to_position_valid_start(self):
        target = Position()
        target.start = True
        target.home = False
        target.safe = None
        target.square = None
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_position(target)
        assert result is position
        assert position == target
        assert "%s" % position == "start"

    def test_move_to_position_valid_home(self):
        target = Position()
        target.start = False
        target.home = True
        target.safe = None
        target.square = None
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_position(target)
        assert result is position
        assert position == target
        assert "%s" % position == "home"

    def test_move_to_position_valid_safe(self):
        target = Position()
        target.start = False
        target.home = False
        target.safe = 3
        target.square = None
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_position(target)
        assert result is position
        assert position == target
        assert "%s" % position == "safe 3"

    def test_move_to_position_valid_square(self):
        target = Position()
        target.start = False
        target.home = False
        target.safe = None
        target.square = 3
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_position(target)
        assert result is position
        assert position == target
        assert "%s" % position == "square 3"

    def test_move_to_position_invalid_multiple(self):
        position = Position()
        for start, home, safe, square in [
            (True, True, None, None),
            (True, False, 1, None),
            (True, False, None, 1),
            (False, True, 1, None),
            (False, True, None, 1),
            (False, False, 1, 1),
        ]:
            with pytest.raises(ValueError):
                target = Position()
                target.start = start
                target.home = home
                target.safe = safe
                target.square = square
                position.move_to_position(target)

    def test_move_to_position_invalid_none(self):
        position = Position()
        with pytest.raises(ValueError):
            target = Position()
            target.start = False
            target.home = False
            target.safe = None
            target.square = None
            position.move_to_position(target)

    def test_move_to_position_invalid_safe(self):
        position = Position()
        for square in [-1000, -2 - 1, 5, 6, 1000]:
            with pytest.raises(ValueError):
                target = Position()
                target.safe = square
                position.move_to_position(target)

    def test_move_to_position_invalid_square(self):
        position = Position()
        for square in [-1000, -2 - 1, 60, 61, 1000]:
            with pytest.raises(ValueError):
                target = Position()
                target.square = square
                position.move_to_position(target)

    def test_move_to_start(self):
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_start()
        assert result is position
        assert position.start is True
        assert position.home is False
        assert position.safe is None
        assert position.square is None
        assert "%s" % position == "start"

    def test_move_to_home(self):
        position = Position()
        position.start = "x"
        position.home = "x"
        position.safe = "x"
        position.square = "x"
        result = position.move_to_home()
        assert result is position
        assert position.start is False
        assert position.home is True
        assert position.safe is None
        assert position.square is None
        assert "%s" % position == "home"

    def test_move_to_safe_valid(self):
        for square in range(SAFE_SQUARES):
            position = Position()
            position.start = "x"
            position.home = "x"
            position.safe = "x"
            position.square = "x"
            result = position.move_to_safe(square)
            assert result is position
            assert position.start is False
            assert position.home is False
            assert position.safe == square
            assert position.square is None
            assert "%s" % position == "safe %d" % square

    def test_move_to_safe_invalid(self):
        for square in [-1000, -2 - 1, 5, 6, 1000]:
            with pytest.raises(ValueError):
                position = Position()
                position.move_to_safe(square)

    def test_move_to_square_valid(self):
        for square in range(BOARD_SQUARES):
            position = Position()
            position.start = "x"
            position.home = "x"
            position.safe = "x"
            position.square = "x"
            result = position.move_to_square(square)
            assert result is position
            assert position.start is False
            assert position.home is False
            assert position.safe is None
            assert position.square is square
            assert "%s" % position == "square %d" % square

    def test_move_to_square_invalid(self):
        for square in [-1000, -2 - 1, 60, 61, 1000]:
            with pytest.raises(ValueError):
                position = Position()
                position.move_to_square(square)


class TestPawn:
    def test_constructor(self):
        pawn = Pawn(PlayerColor.RED, 0)
        assert pawn.color == PlayerColor.RED
        assert pawn.index == 0
        assert pawn.name == "Red0"
        assert pawn.position == Position()
        assert "%s" % pawn == "Red0->start"  # because default position is in start

    def test_constructor_with_name(self):
        pawn = Pawn(PlayerColor.RED, 0, name="whatever")
        assert pawn.color == PlayerColor.RED
        assert pawn.index == 0
        assert pawn.name == "whatever"
        assert pawn.position == Position()
        assert "%s" % pawn == "whatever->start"  # because default position is in start


class TestPlayer:
    def test_constructor(self):
        player = Player(PlayerColor.RED)
        assert player.color == PlayerColor.RED
        assert player.turns == 0
        assert len(player.pawns) == PAWNS
        for pawn in player.pawns:
            assert isinstance(pawn, Pawn)

    def test_copy(self):
        player = Player(PlayerColor.RED)
        player.pawns[0].position.move_to_home()
        player.pawns[1].position.move_to_safe(2)
        player.pawns[2].position.move_to_square(32)
        copy = player.copy()
        assert copy is not player and copy == player

    def test_find_first_pawn_in_start(self):
        player = Player(PlayerColor.RED)
        for i in range(PAWNS):
            assert player.find_first_pawn_in_start() is player.pawns[i]
            player.pawns[i].position.move_to_home()
        assert player.find_first_pawn_in_start() is None

    def test_all_pawns_in_home(self):
        player = Player(PlayerColor.RED)
        for i in range(PAWNS):
            assert player.all_pawns_in_home() is False
            player.pawns[i].position.move_to_home()
        assert player.all_pawns_in_home() is True


class TestHistory:
    def test_constructor(self):
        color = PlayerColor.BLUE
        card = CardType.CARD_APOLOGIES
        history = History("action", color, card)
        assert history.action == "action"
        assert history.color is color
        assert history.card is card
        assert history.timestamp <= DateTime.utcnow()

    def test_str(self):
        timestamp = pendulum.parse("2020-03-25T14:02:16")

        history = History("This is an action", color=None, timestamp=timestamp)
        assert "%s" % history == "[14:02:16] General - This is an action"

        history = History("This is an action", color=PlayerColor.BLUE, timestamp=timestamp)
        assert "%s" % history == "[14:02:16] Blue - This is an action"

        history = History("This is an action", color=PlayerColor.BLUE, card=CardType.CARD_10, timestamp=timestamp)
        assert "%s" % history == "[14:02:16] Blue - This is an action"


class TestPlayerView:
    def test_constructor(self):
        player = Player(PlayerColor.RED)
        opponents = {PlayerColor.GREEN: Player(PlayerColor.GREEN)}
        view = PlayerView(player, opponents)
        assert view.player == player
        assert view.opponents == opponents

    def test_copy(self):
        player = Player(PlayerColor.RED)
        opponents = {PlayerColor.GREEN: Player(PlayerColor.GREEN)}
        view = PlayerView(player, opponents)
        copy = view.copy()
        assert copy is not view and copy == view

    def test_get_pawn(self):
        player = Player(PlayerColor.RED)
        opponents = {PlayerColor.GREEN: Player(PlayerColor.GREEN)}
        view = PlayerView(player, opponents)
        assert view.get_pawn(MagicMock(color=PlayerColor.RED, index=3)) == view.player.pawns[3]
        assert view.get_pawn(MagicMock(color=PlayerColor.GREEN, index=1)) == view.opponents[PlayerColor.GREEN].pawns[1]
        assert view.get_pawn(MagicMock(color=PlayerColor.YELLOW, index=0)) is None

    def test_all_pawns(self):
        player = Player(PlayerColor.RED)
        opponents = {PlayerColor.GREEN: Player(PlayerColor.GREEN)}
        view = PlayerView(player, opponents)
        pawns = view.all_pawns()
        assert len(pawns) == 2 * PAWNS
        for i in range(PAWNS):
            assert player.pawns[i] in pawns
            assert opponents[PlayerColor.GREEN].pawns[i] in pawns


class TestGame:
    def test_constructor_2_players_standard(self):
        game = Game(2)
        assert len(game.players) == 2
        assert len(game.history) == 0
        for color in [PlayerColor.RED, PlayerColor.YELLOW]:
            assert game.players[color].color == color
            assert len(game.players[color].hand) == 0
        assert game.deck is not None

    def test_constructor_3_players_standard(self):
        game = Game(3)
        assert len(game.players) == 3
        assert len(game.history) == 0
        for color in [PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN]:
            assert game.players[color].color == color
            assert len(game.players[color].hand) == 0
        assert game.deck is not None

    def test_constructor_4_players_standard(self):
        game = Game(4)
        assert len(game.players) == 4
        assert len(game.history) == 0
        for color in [PlayerColor.RED, PlayerColor.YELLOW, PlayerColor.GREEN, PlayerColor.BLUE]:
            assert game.players[color].color == color
            assert len(game.players[color].hand) == 0
        assert game.deck is not None

    def test_constructor_invalid_players(self):
        for playercount in [-2, -1, 0, 1, 5, 6]:
            with pytest.raises(ValueError):
                Game(playercount)

    def test_started(self):
        game = Game(4)
        assert game.started is False
        game.track("whatever")
        assert game.started is True

    def test_completed_and_winner(self):
        game = Game(4)

        # move all but last pawn into home for all of the players; the game is not complete
        for player in game.players.values():
            for i in range(PAWNS - 1):
                assert game.completed is False
                player.pawns[i].position.move_to_home()

        # move the final pawn to home for one player; now the game is complete
        game.players[PlayerColor.RED].pawns[PAWNS - 1].position.move_to_home()

        assert game.completed is True
        assert game.winner is game.players[PlayerColor.RED]

    def test_copy(self):
        game = TestGame._create_realistic_game()
        copy = game.copy()
        assert copy == game

    def test_json_roundtrip(self):
        game = TestGame._create_realistic_game()
        data = game.to_json()
        copy = Game.from_json(data)
        assert copy == game

    def test_track_no_player(self):
        game = Game(4)
        game.track("action")
        assert game.history[0].action == "action"
        assert game.history[0].color is None
        assert game.history[0].card is None
        assert game.history[0].timestamp <= DateTime.utcnow()
        assert game.players[PlayerColor.RED].turns == 0
        assert game.players[PlayerColor.YELLOW].turns == 0
        assert game.players[PlayerColor.BLUE].turns == 0
        assert game.players[PlayerColor.GREEN].turns == 0

    def test_track_with_color(self):
        game = Game(4)
        player = MagicMock(color=PlayerColor.RED)
        card = MagicMock(cardtype=CardType.CARD_12)
        game.track("action", player, card)
        assert game.history[0].action == "action"
        assert game.history[0].color is PlayerColor.RED
        assert game.history[0].card == CardType.CARD_12
        assert game.history[0].timestamp <= DateTime.utcnow()
        assert game.players[PlayerColor.RED].turns == 1
        assert game.players[PlayerColor.YELLOW].turns == 0
        assert game.players[PlayerColor.BLUE].turns == 0
        assert game.players[PlayerColor.GREEN].turns == 0

    def test_create_player_view_invalid(self):
        game = Game(2)
        with pytest.raises(KeyError):
            game.create_player_view(PlayerColor.BLUE)  # no blue player in 2-player game

    def test_create_player_view(self):
        game = Game(4)

        game.players[PlayerColor.RED].hand.append(game.deck.draw())
        game.players[PlayerColor.YELLOW].hand.append(game.deck.draw())
        game.players[PlayerColor.GREEN].hand.append(game.deck.draw())
        game.players[PlayerColor.BLUE].hand.append(game.deck.draw())

        view = game.create_player_view(PlayerColor.RED)

        assert game.players[PlayerColor.RED] is not view.player
        assert game.players[PlayerColor.YELLOW] is not view.opponents[PlayerColor.YELLOW]

        assert game.players[PlayerColor.RED] == view.player

        for color in [PlayerColor.YELLOW, PlayerColor.GREEN, PlayerColor.BLUE]:
            assert view.opponents[color].color == color
            assert len(view.opponents[color].hand) == 0
            assert view.opponents[color].pawns == game.players[color].pawns

    @staticmethod
    def _create_realistic_game():
        """Create a realistic game with changes to the defaults for all types of values."""
        game = Game(4)
        game.track("this happened")
        game.track("another thing", game.players[PlayerColor.RED])
        card1 = game.deck.draw()
        card2 = game.deck.draw()
        game.deck.draw()  # just throw it away
        game.deck.discard(card1)
        game.deck.discard(card2)
        game.players[PlayerColor.RED].pawns[0].position.move_to_square(32)
        game.players[PlayerColor.BLUE].pawns[2].position.move_to_home()
        game.players[PlayerColor.BLUE].hand.append(card1)
        game.players[PlayerColor.YELLOW].pawns[3].position.move_to_safe(1)
        game.players[PlayerColor.GREEN].pawns[1].position.move_to_square(19)
        game.players[PlayerColor.GREEN].hand.append(card2)
        return game



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

"""
Implements a quick'n'dirty game-playing demo using curses.
"""

import platform

from .game import GameMode
from .source import CharacterInputSource

if platform.system() == "Windows":
    from .demo_windows import run_demo as implementation  # pylint: disable=unused-import
else:
    from .demo_unix import run_demo as implementation  # pylint: disable=unused-import


def run_demo(players: int, mode: GameMode, source: CharacterInputSource, delay_sec: float, exit_immediately: bool) -> None:
    """
    Run the quick'n'dirty demo in a terminal window.

    Args:
        players(int): Number of players in the game
        mode(GameMode): The game mode
        source(CharacterInputSource): The source to use for choosing player moves
        delay_sec(float): The delay between turns when executing the game
        exit_immediately(bool): Exit immediately when the game completes
    """
    implementation(players, mode, source, delay_sec, exit_immediately)



# -*- coding: utf-8 -*-
# vim: set ft=python ts=4 sw=4 expandtab:

import pytest

from apologies.util import CircularQueue


class TestCircularQueue:
    def test_constructor_empty(self):
        with pytest.raises(ValueError):
            CircularQueue([])

    def test_constructor_single(self):
        queue = CircularQueue(["a"])
        assert queue.entries == ["a"]
        assert queue.next() == "a"
        assert queue.next() == "a"
        assert queue.next() == "a"
        assert queue.next() == "a"
        assert queue.next() == "a"
        assert queue.next() == "a"

    def test_constructor_single_none(self):
        queue = CircularQueue([None])
        assert queue.entries == [None]
        assert queue.next() is None
        assert queue.next() is None
        assert queue.next() is None
        assert queue.next() is None
        assert queue.next() is None
        assert queue.next() is None

    def test_constructor_multiple(self):
        queue = CircularQueue(["a", "b", "c", "d", "e", None])
        assert queue.entries == ["a", "b", "c", "d", "e", None]
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None
        assert queue.next() == "a"

    def test_constructor_multiple_first(self):
        queue = CircularQueue(["a", "b", "c", "d", "e", None], first="c")
        assert queue.entries == ["a", "b", "c", "d", "e", None]
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"

    def test_constructor_multiple_first_none(self):
        queue = CircularQueue(["a", "b", "c", "d", "e", None], first=None)
        assert queue.entries == ["a", "b", "c", "d", "e", None]
        assert queue.next() is None
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None
        assert queue.next() == "a"
        assert queue.next() == "b"
        assert queue.next() == "c"
        assert queue.next() == "d"
        assert queue.next() == "e"
        assert queue.next() is None

    def test_constructor_first_invalid(self):
        with pytest.raises(ValueError):
            CircularQueue([], first=None)
        with pytest.raises(ValueError):
            CircularQueue([], first="")
        with pytest.raises(ValueError):
            CircularQueue([], first="f")
        with pytest.raises(ValueError):
            CircularQueue(["a"], first=None)
        with pytest.raises(ValueError):
            CircularQueue(["a"], first="")
        with pytest.raises(ValueError):
            CircularQueue(["a"], first="f")
        with pytest.raises(ValueError):
            CircularQueue(["a", "b", "c", "d", "e"], first=None)
        with pytest.raises(ValueError):
            CircularQueue(["a", "b", "c", "d", "e"], first="")
        with pytest.raises(ValueError):
            CircularQueue(["a", "b", "c", "d", "e"], first="f")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['apologies']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.2.0,<23.0.0', 'cattrs>=22.2.0,<23.0.0', 'pendulum>=2.1.2,<3.0.0']

extras_require = \
{'docs': ['importlib-metadata>=6.0.0,<7.0.0',
          'sphinx>=6.1.3,<7.0.0',
          'sphinx-autoapi>=2.0.1,<3.0.0']}

setup_kwargs = {
    'name': 'apologies',
    'version': '0.1.43',
    'description': 'Python code to play a game similar to Sorry',
    'long_description': '# Apologies Python Library\n\n[![pypi](https://img.shields.io/pypi/v/apologies.svg)](https://pypi.org/project/apologies/)\n[![license](https://img.shields.io/pypi/l/apologies.svg)](https://github.com/pronovic/apologies/blob/master/LICENSE)\n[![wheel](https://img.shields.io/pypi/wheel/apologies.svg)](https://pypi.org/project/apologies/)\n[![python](https://img.shields.io/pypi/pyversions/apologies.svg)](https://pypi.org/project/apologies/)\n[![Test Suite](https://github.com/pronovic/apologies/workflows/Test%20Suite/badge.svg)](https://github.com/pronovic/apologies/actions?query=workflow%3A%22Test+Suite%22)\n[![docs](https://readthedocs.org/projects/apologies/badge/?version=stable&style=flat)](https://apologies.readthedocs.io/en/stable/)\n[![coverage](https://coveralls.io/repos/github/pronovic/apologies/badge.svg?branch=master)](https://coveralls.io/github/pronovic/apologies?branch=master)\n\n[Apologies](https://github.com/pronovic/apologies) is a Python library that implements a game similar to the [Sorry](https://en.wikipedia.org/wiki/Sorry!_(game)) board game.  On UNIX-like platforms, it includes a console demo that plays the game with automated players, intended for use by developers and not by end users.  See the [documentation](https://apologies.readthedocs.io/en/stable) for notes about the public interface.\n\nIt was written as a learning exercise and technology demonstration effort, and\nserves as a complete example of how to manage a modern (circa 2020) Python\nproject, including style checks, code formatting, integration with IntelliJ, CI\nbuilds at GitHub, and integration with PyPI and Read the Docs.\n',
    'author': 'Kenneth J. Pronovici',
    'author_email': 'pronovic@ieee.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/apologies/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)
